/*
 *    ServiceCluster.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// leave the package name as is for now - should be changed to something better
// cluster? groups?
package org.greenstone.gsdl3.collection;


import org.greenstone.gsdl3.util.*;
import org.greenstone.gsdl3.core.*;
import org.greenstone.gsdl3.service.*;

// java XML classes we're using
import org.w3c.dom.Document; 
import org.w3c.dom.Node; 
import org.w3c.dom.Element; 
import org.w3c.dom.NodeList; 

import java.io.File;
import java.util.HashMap;

/* ServiceCluster - a groups of services that are related in some way
 * Implements ModuleInterface. Contains a list of services provided by the cluster, along with metadata about the cluster itself.
 * a collection is a special type of cluster
 *  @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 *  @version $Revision: 3512 $
 *  @see ModuleInterface
 */
public class ServiceCluster  
    implements ModuleInterface {


    /** base directory for the site that this cluster belongs to*/
    protected String site_home_ = null;

    /** The name of the cluster - for a collection, this is the collection name*/
    protected String cluster_name_ = null;

    /** a reference to the message router */
    protected MessageRouter router_ = null;
    /** The map of services.
     *
     * Maps Services to ServicesImpl objects
     * @see ServicesImpl
     *
     */
    protected HashMap service_map_=null;

    /** XML converter for String to DOM and vice versa */
    protected XMLConverter converter_=null;

    /** container doc for description elements */
    protected Document doc_ = null;
    /** list of services */
    protected Element service_info_ = null;
    /** list of Metadata */
    protected Element meta_info_ = null;
    /** full description */
    protected Element description_ = null;

    public void setSiteHome(String home) {
	site_home_ = home;
    }

    public void setClusterName(String name) {
	cluster_name_ = name;	
	description_.setAttribute(GSXML.NAME_ATT, name);
    }  

    public void setMessageRouter(MessageRouter m) {
	router_ = m;
    }

    public ServiceCluster() {
	service_map_ = new HashMap();
	converter_ = new XMLConverter();
	doc_ = converter_.newDOM();
	description_ = doc_.createElement(GSXML.CLUSTER_ELEM);
	service_info_ = doc_.createElement(GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER);
	meta_info_ = doc_.createElement(GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	
    }

    /**
     * Configures the cluster.
     *
     * gsdlHome and clusterName must be set before configure is called.
     *
     * reads the site configuration file, and configures itself
     * this calls configure(Element) with the XML element node from the config
     * file.
     * configure(Element) should be used if the config file has already been 
     * parsed.  This method will work with any subclass.
     * 
     * @return true if configure successful, false otherwise.
     */
    public boolean configure() {

	if (site_home_ == null || cluster_name_== null) {
	    System.err.println("site_home_ and cluster_name_ must be set before configure called!");
	    return false;
	}
	System.out.println("configuring service cluster");
	// read the site configuration file
	File config_file_ = new File(GSFile.siteConfigFile(site_home_));

	if (!config_file_.exists()) {
	    System.err.println(config_file_+" does not exist");
	    System.err.println("couldn't configure cluster: "+cluster_name_);
	    return false;
	}

	Document doc = converter_.getDOM(config_file_);
	
	// get the appropriate service cluster element
	Element cluster_list = (Element)GSXML.getChildByTagName(doc.getDocumentElement(), GSXML.CLUSTER_ELEM+GSXML.LIST_MODIFIER);
	Element sc = GSXML.getNamedElement(cluster_list, GSXML.CLUSTER_ELEM,
					   GSXML.NAME_ATT, cluster_name_);
	
	return this.configure(sc);
    }
    
    public boolean configure(Element cluster_info) {
	
	// create all the services
	NodeList nodes = cluster_info.getElementsByTagName(GSXML.SERVICE_IMPL_ELEM);
	if (nodes.getLength()==0) {
	    System.err.println("Cluster configuration error: cluster "+cluster_name_+" has no service modules!");
	    return false;
	}

	for(int i=0; i<nodes.getLength(); i++) {
	
	    // the xml request to send to the servicesImpl to query what
	    // services it provides
	    Element message = doc_.createElement(GSXML.MESSAGE_ELEM);
	    Element request = doc_.createElement(GSXML.REQUEST_ELEM);
	    message.appendChild(request);
	    request.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);
	    request.setAttribute(GSXML.INFO_ATT, GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER);
	    

	    Element n = (Element)nodes.item(i);
	    String servicetype = n.getAttribute(GSXML.NAME_ATT);
	    
	    try {
		ServicesImpl s = (ServicesImpl)Class.forName("org.greenstone.gsdl3.service."+servicetype).newInstance();

		s.setSiteHome(site_home_);
		s.setClusterName(cluster_name_);
		s.setMessageRouter(router_);
		// pass the xml node to the service for configuration
		s.configure(n);
		
		// find out the supported service types for this service module
		Node types = s.process(message);
		NodeList typenodes = ((Element)types).getElementsByTagName(GSXML.SERVICE_ELEM);	   
		
		for (int j=0; j<typenodes.getLength();j++) {
		    String service = ((Element) typenodes.item(j)).getAttribute(GSXML.NAME_ATT);	    
		    service_map_.put(service, s);
		    
		    // also add info to the ServiceInfo XML element
		    service_info_.appendChild(doc_.importNode(typenodes.item(j), true));
		}
	    } catch (Exception e) {
		System.out.println("Cluster configure exception: couldn't create service module:org.greenstone.gsdl3.service."+servicetype+"\n"+e.getMessage() + e.getClass() );
		e.printStackTrace();
		//return false;
	    }
	    
	}
	
	// get the metadata - for now just add it to the list
	Element meta_list = (Element) GSXML.getChildByTagName(cluster_info, GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	NodeList metanodes = meta_list.getElementsByTagName(GSXML.METADATA_ELEM);
	
	if (metanodes.getLength()>0) { 	
	    for(int k=0; k<metanodes.getLength(); k++) {
		meta_info_.appendChild(doc_.importNode(metanodes.item(k),true));
	    }
	}	     
    
	// add elements to description
	description_.appendChild(service_info_);
	description_.appendChild(meta_info_);
	return true;
	
	
    }

    /**
     * Process an XML document - uses Strings 
     *  just calls process(Node).
     *
     * @param in the Document to process - a string
     * @return the resultant document as a string - contains any error messages
     * @see String
     */
    public String process(String in) {

	Document doc = converter_.getDOM(in);
	Element e = doc.getDocumentElement();
	
	Element res = process(e);
	return converter_.getString(res);
	
    }
    
    /** process XML as Node
     *
     */
    public Element process(Element message) {

	NodeList requests = message.getElementsByTagName(GSXML.REQUEST_ELEM);
	Document mess_doc = message.getOwnerDocument();
	Element mainResult = doc_.createElement(GSXML.MESSAGE_ELEM);
	if (requests.getLength()==0) {
	    System.out.println("no requests for cluster:"+cluster_name_);
	    // no requests
	    return mainResult; // for now
	}
	for (int i=0; i<requests.getLength(); i++) {
	    Element request = (Element)requests.item(i);
	    String to = request.getAttribute(GSXML.TO_ATT);

	    // the cluster name should be first, check, then remove
	    String clustername = GSPath.getFirstLink(to);
	    if (!clustername.equals(cluster_name_)){ 
		System.out.println("ServiceCluster error: cluster name wrong! was "+clustername+" should have been "+cluster_name_);
		continue; // ignore this request
	    }
	    to = GSPath.removeFirstLink(to);
	    request.setAttribute(GSXML.TO_ATT, to);
	   
	    if (to.equals("")) { // this command is for me
		String type = request.getAttribute(GSXML.TYPE_ATT);
		if (type.equals(GSXML.REQUEST_TYPE_DESCRIBE)) {
		    // create a response with the description inside it
		    Element response = doc_.createElement(GSXML.RESPONSE_ELEM);
		    response.setAttribute(GSXML.FROM_ATT, cluster_name_);
		    response.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);
		    response.appendChild(description_);
		    mainResult.appendChild(response);
		
		} else {
		    System.out.println("ServiceCluster: error, cant handle request:");
		    System.out.println(converter_.getString(request));
		    
		}
	    
	    } else { // the request is for one of my services
		String service = GSPath.getFirstLink(to);
		
		if (!service_map_.containsKey(service)) {
		    System.err.println("ServiceCluster: non-existant service, "+service+", specified, in:\n"+converter_.getString(request));
		    continue;
		}
		// have to pass the request to the service
		Element single_message = mess_doc.createElement(GSXML.MESSAGE_ELEM);
		single_message.appendChild(request);
		Element response = ((ModuleInterface)service_map_.get(service)).process(single_message);
		if (response != null) {
		    mainResult.appendChild(doc_.importNode(GSXML.getChildByTagName(response, "response"), true));
		}
		  
	    } // else
		
	   
	} // for each request
	System.out.println("servicecluster returning result");
	System.out.println(converter_.getString(mainResult));
	return mainResult;
    }
	
}






